#define MODULE "main"

#include "acpi/RSDT.h"
#include "assert.h"
#include "config.h"
#include "cpu/CPU.h"
#include "gdt/GDT.h"
#include "init/Init.h"
#include "init/InitRD.h"
#include "interrupts/IDT.h"
#include "interrupts/Install.h"
#include "interrupts/Interrupts.h"
#include "io/PCI.h"
#include "io/PIC.h"
#include "io/Serial.h"
#include "log/Log.h"
#include "memory/Memory.h"
#include "memory/MemoryManager.h"
#include "memory/MemoryMap.h"
#include "memory/PMM.h"
#include "memory/VMM.h"
#include "misc/PCITypes.h"
#include "misc/reboot.h"
#include "panic/Panic.h"
#include "rand/Mersenne.h"
#include "render/Framebuffer.h"
#include "render/TextRenderer.h"
#include "std/stdio.h"
#include "std/stdlib.h"
#include "std/string.h"
#include "sys/elf/ELFLoader.h"
#include "thread/PIT.h"
#include "thread/Scheduler.h"

extern "C" void _userspace();

extern "C" void _start()
{
    Init::check_magic();
    Init::disable_smp(); // Put all other cores except the bootstrap one in an infinite loop
    Init::early_init();

    kinfoln("Starting Moon %s", moon_version());

    CPU::log_cpu_information();

    Memory::walk_memory_map();

    GDT::load();

    kinfoln("Loaded GDT");

    Interrupts::install();

    IDT::load();

    kinfoln("Loaded IDT");

    PIC::remap();
    PIC::enable_master(0b11111100); // enable keyboard and PIT
    PIC::enable_slave(0b11111111);

    kinfoln("Prepared PIC");

    PIT::initialize(1000); // 1000 times per second

    kinfoln("Prepared PIT");

    Scheduler::init();

    kinfoln("Prepared scheduler");

    Scheduler::add_kernel_task([]() {
        int64_t x = 0;
        int64_t y = 0;
        int64_t xvel = 10;
        int64_t yvel = 10;
        while (1)
        {
            sleep(2);
            uint32_t color = (uint32_t)Mersenne::get();
            uint32_t* colptr = &color;
            x += xvel;
            y += yvel;
            if ((x + 10) >= framebuffer0.width())
            {
                xvel = -xvel;
                x = (framebuffer0.width() - 10);
            }
            if ((y + 10) >= framebuffer0.height())
            {
                yvel = -yvel;
                y = (framebuffer0.height() - 10);
            }
            if (xvel < 0 && (x - 10) < 0)
            {
                xvel = -xvel;
                x = 0;
            }
            if (yvel < 0 && (y - 10) < 0)
            {
                yvel = -yvel;
                y = 0;
            }
            framebuffer0.paint_rect(x, y, 10, 10, *(Color*)colptr);
        }
    });

    Scheduler::add_kernel_task([]() {
        while (1)
        {
            sleep(100);
            uint32_t color = (uint32_t)Mersenne::get();
            uint32_t* colptr = &color;
            framebuffer0.paint_rect(Mersenne::get() % (framebuffer0.width() - 256),
                                    Mersenne::get() % (framebuffer0.height() - 256), Mersenne::get() % 255,
                                    Mersenne::get() % 255, *(Color*)colptr);
        }
    });

    Scheduler::add_kernel_task([]() {
        while (1)
        {
            sleep(400);
            Scheduler::reap_tasks();
        }
    });

    void* elf_start_address = ELFLoader::load_elf_from_initrd("bin/init");
    if (!elf_start_address) kerrorln("failed to load /bin/init from initrd");
    else
    {
        kinfoln("loaded /bin/init at %lx", (uint64_t)elf_start_address);
        Scheduler::add_user_task(elf_start_address);
    }

    kinfoln("Prepared scheduler tasks");

    framebuffer0.clear(Color::Cyan);

    Interrupts::enable();

    kinfoln("Interrupts enabled");

    PCI::scan([](PCI::Device& dev) {
        kinfoln("Found PCI device %x:%x, %s", dev.id().vendor, dev.id().device, pci_type_name(dev.type()));
    });

    Scheduler::exit();
}