#define MODULE "panic"

#include "panic/Panic.h"
#include "init/InitRD.h"
#include "interrupts/IDT.h"
#include "io/PIC.h"
#include "log/Log.h"
#include "misc/MSR.h"
#include "std/stdio.h"
#include "thread/Scheduler.h"
#include "trace/StackTracer.h"

static void dump_registers(Context* context)
{
    printf("-- Registers: \n");
    printf("rax: %lx, rbx: %lx, rcx: %lx, rdx: %lx\n", context->rax, context->rbx, context->rcx, context->rdx);
    printf("rsi: %lx, rdi: %lx, rsp: %lx, rbp: %lx\n", context->rsi, context->rdi, context->rsp, context->rbp);
    printf("r8: %lx, r9: %lx, r10: %lx, r11: %lx\n", context->r8, context->r9, context->r10, context->r11);
    printf("r12: %lx, r13: %lx, r14: %lx, r15: %lx\n", context->r12, context->r13, context->r14, context->r15);
    printf("rip: %lx, cs: %lx, ss: %lx\n", context->rip, context->cs, context->ss);
    printf("rflags: %lx, cr2: %lx\n", context->rflags, context->cr2);
    printf("ia32_efer: %lx\n", MSR::read_from(IA32_EFER_MSR));
}

[[noreturn]] static void __panic_stub(Context* context)
{
    dump_registers(context);

    if (InitRD::is_initialized())
    {
        printf("-- Stack trace:\n");
        StackTracer tracer(context->rbp);
        tracer.trace_with_ip(context->rip);
    }
    else { printf("-- No stack trace available\n"); }

    PIC::enable_master(0b11111101); // enable keyboard only
    PIC::enable_slave(0b11111111);

    IDTR idtr;
    idtr.limit = 0x0000;
    idtr.offset = 0x0000;
    asm volatile(
        "lidt %0"
        :
        : "m"(
            idtr)); // when an interrupt arrives, triple-fault (the only interrupt that should come is the keyboard one)

    asm volatile("sti");

    kinfoln("Press any key to restart.");

    while (1) asm volatile("hlt");
}

extern "C" [[noreturn]] void __do_int_panic(Context* context, const char* file, int line, const char* message)
{
    asm volatile("cli");

    if (context->number >= 0x20 && context->number < 0x30) { PIC::send_eoi(context->irq_number); }

    Task* task;
    if ((task = Scheduler::current_task()))
    {
        kerrorln("Kernel panic in task %ld, at %s, line %d: %s", task->id, file, line, message);
    }
    else { kerrorln("Kernel panic at %s, line %d: %s", file, line, message); }

    __panic_stub(context);
}

extern "C" [[noreturn]] void __do_panic(Context* context, const char* message)
{
    Task* task;
    if ((task = Scheduler::current_task())) { kerrorln("Kernel panic in task %ld: %s", task->id, message); }
    else { kerrorln("Kernel panic: %s", message); }

    __panic_stub(context);
}